@@ -4,7 +4,7 @@ from dataclasses import dataclass, field, fields
 from time import monotonic
 from typing import Type, Iterable, NamedTuple, Union, Iterator, Optional
 
-from clickhouse_driver import Client
+from clickhouse_pool import ChPool
 from decouple import config
 
 from hcube.api.backend import CubeBackend
@@ -74,17 +74,25 @@ class ClickhouseCubeBackend(CubeBackend):
         (FloatMetric, "Float"),
         (IntMetric, "Int"),
     )
+    default_settings = {
+        "aggregate_functions_null_for_empty": 1,
+        "allow_experimental_projection_optimization": 1,
+    }
 
     def __init__(
         self,
         database=None,
+        query_settings=None,
         **client_attrs,
     ):
         super().__init__()
         self.database = database
         assert self.database, "database must be present"
+        self.query_settings = {**self.default_settings}
+        if query_settings:
+            self.query_settings.update(query_settings)
         self.client_attrs = client_attrs
-        self.client = Client(database=database, **client_attrs)
+        self.pool = ChPool(database=database, **client_attrs)
         self._table_exists = {}
 
     def create_table(self, cube: Type[Cube]):
@@ -92,18 +100,18 @@ class ClickhouseCubeBackend(CubeBackend):
         if not self._table_exists.get(table_name, False):
             self._init_table(cube)
             self._create_materialized_views(cube)
-            self.client.execute("SET allow_experimental_projection_optimization = 1;")
             self._table_exists[table_name] = True
 
     def store_records(self, cube: Type[Cube], records: Iterable[NamedTuple]):
-        self.client.execute(f"USE {self.database}")
-        self.create_table(cube)
-        meta = self.get_table_meta(cube)
-        clean_records = cube.cleanup_records(records)
-        self.client.execute(
-            f"INSERT INTO {self.database}.{self._cube_to_table_name(cube)} VALUES ",
-            ({**rec._asdict(), meta.sign_col: 1} for rec in clean_records),
-        )
+        with self.pool.get_client() as client:
+            client.execute(f"USE {self.database}")
+            self.create_table(cube)
+            meta = self.get_table_meta(cube)
+            clean_records = cube.cleanup_records(records)
+            client.execute(
+                f"INSERT INTO {self.database}.{self._cube_to_table_name(cube)} VALUES ",
+                ({**rec._asdict(), meta.sign_col: 1} for rec in clean_records),
+            )
 
     def get_records(
         self,
@@ -128,10 +136,11 @@ class ClickhouseCubeBackend(CubeBackend):
         # we could use execute_iter here, but is breaks the communication if some stuff is left
         # unconsumed "in the wire", so this seems safer
         start = monotonic()
-        output = self.client.execute(text, params)
-        logger.debug(f"Query time: {monotonic() - start: .3f} s")
-        for rec in output:
-            yield result(*rec)
+        with self.pool.get_client() as client:
+            output = client.execute(text, params)
+            logger.debug(f"Query time: {monotonic() - start: .3f} s")
+            for rec in output:
+                yield result(*rec)
 
     def delete_records(self, query: CubeQuery) -> None:
         """
@@ -156,8 +165,9 @@ class ClickhouseCubeBackend(CubeBackend):
         whole_text = f"INSERT INTO {table} ({field_names}, {meta.sign_col}) {text}"
         logger.debug("Delete query: %s, params: %s", whole_text, params)
         start = monotonic()
-        self.client.execute(whole_text, params)
-        logger.debug(f"Query time: {monotonic() - start: .3f} s")
+        with self.pool.get_client() as client:
+            client.execute(whole_text, params)
+            logger.debug(f"Query time: {monotonic() - start: .3f} s")
 
     @classmethod
     def get_table_meta(cls, cube: Type[Cube]) -> TableMetaParams:
@@ -236,7 +246,9 @@ class ClickhouseCubeBackend(CubeBackend):
         # ensure MAX([]) is None not 0
         # but this does not seem to be compatible with automatic use of projections, so we
         # might need to turn it off in the future
-        text += " SETTINGS aggregate_functions_null_for_empty = 1"
+        settings_part = ", ".join(f"{k} = {v}" for k, v in self.default_settings.items())
+        if settings_part:
+            text += f" SETTINGS {settings_part}"
         return text, params, fields, matview
 
     def _translate_aggregation(self, agg: Aggregation, sign_column: Optional[str]) -> str:
@@ -337,7 +349,8 @@ class ClickhouseCubeBackend(CubeBackend):
             f"{settings_part};"
         )
         logger.debug(command)
-        self.client.execute(command)
+        with self.pool.get_client() as client:
+            client.execute(command)
 
     def _ch_type(self, dimension: Union[Dimension, Metric]):
         for dim_cls, ch_type in self.dimension_type_map:
@@ -378,7 +391,8 @@ class ClickhouseCubeBackend(CubeBackend):
             f"GROUP BY {preserved};"
         )
         logger.debug(command)
-        self.client.execute(command)
+        with self.pool.get_client() as client:
+            client.execute(command)
 
     def _create_projection(
         self, cube: Type[Cube], matview: Type[AggregatingMaterializedView], populate=True
@@ -396,11 +410,12 @@ class ClickhouseCubeBackend(CubeBackend):
             f"(SELECT {preserved}, {agg_part} GROUP BY {preserved});"
         )
         logger.debug(command)
-        self.client.execute(command)
-        if populate:
-            self.client.execute(
-                f"ALTER TABLE {self.database}.{table_name} MATERIALIZE PROJECTION {view_name}"
-            )
+        with self.pool.get_client() as client:
+            client.execute(command)
+            if populate:
+                client.execute(
+                    f"ALTER TABLE {self.database}.{table_name} MATERIALIZE PROJECTION {view_name}"
+                )
 
     def _translate_transform(self, transform: Transform) -> (str, str):
         """
