@@ -22,6 +22,7 @@ from hcube.api.models.filters import (
     IsNullFilter,
     ComparisonFilter,
     NegativeListFilter,
+    EqualityFilter,
 )
 from hcube.api.models.metrics import FloatMetric, Metric, IntMetric
 from hcube.api.models.query import CubeQuery
@@ -195,6 +196,8 @@ class PostgresCubeBackend(CubeBackend):
             return f"{fltr.dimension.name} IS{modifier} NULL", {}
         if isinstance(fltr, ComparisonFilter):
             return f"{fltr.dimension.name} {fltr.comparison.value} %({key})s", {key: fltr.value}
+        if isinstance(fltr, EqualityFilter):
+            return f"{fltr.dimension.name} = %({key})s", {key: fltr.value}
         raise ValueError(f"unsupported filter {fltr.__class__}")
 
     def _agg_name(self, agg: AggregationOp):
